# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import os
import sys
from sys import exit
import threading
import signal
import socket
import logging
import time
from Crypto.PublicKey import RSA
import datetime
from struct import pack, unpack
import json
import random

from configuration import get_configuration_value
from socket_communication import SecureCommunicationListen
import database_query as dq
import updater as ud
import rsa_helper as rh
import locator
from helpers import human_readable_file_size


"""
the listener serves encrypted data to anyone knowing the name of the data
It stores data in a SQLite database
"""

# Set global flogging:
logging.basicConfig(
     filename='the.log',
     level=logging.INFO,
     format='[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
     datefmt='%H:%M:%S'
)
# Set up logging to console
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')  # set simple console format
console.setFormatter(formatter)
logging.getLogger('').addHandler(console)  # add the handler to the root logger

logger = logging.getLogger(__name__)
# Logging setup done

# Global shut down boolean
force_shutdown = False


class Listen:
    def __init__(self):
        self.listensocket = None
        self.clientsocket = None

        print('Allocated storage space for other nodes:' + human_readable_file_size(get_configuration_value('max_storage_space_other_mb')*1024*1024))
        print('Max filesize for other nodes:' + human_readable_file_size(get_configuration_value('swicharoo_file_maxsize_kb')*1024))
        print('Always accept files under this size (if we have storage space):' + human_readable_file_size(get_configuration_value('swicharoo_minsize_accept_all')))

    def accept_swicharoo(self, my_file_size, their_file_size):
        # Check we have the place to store the file:
        used_space = dq.get_shared_files_size()
        if used_space + their_file_size > get_configuration_value('max_storage_space_other_mb')*1024*1024:
            # print('Rejected swicharoo because we are full: used='+ str(used_space) + ' their size='+str(their_file_size) + ' storage space=' + str(get_configuration_value('max_storage_space_other_mb')*1024*1024) )
            logging.info("Listener: Swicharoo denied because we are full")
            return False

        if their_file_size > get_configuration_value('swicharoo_file_maxsize_kb')*1024:
            # print('Rejected swicharoo because their file was too big')
            logging.info("Listener: Swicharoo denied because their file is too big")
            return False

        # Auto accept?
        if their_file_size <= get_configuration_value('swicharoo_minsize_accept_all') or get_configuration_value('swicharoo_policy') == "GRANT_ALL":
            return True

        if get_configuration_value('swicharoo_policy') == "MAX_DOUBLE":
            logging.info("Listener: Swicharoo denied because their file is more than twice ours")
            return their_file_size <= my_file_size*2

        return True

    def forcestop(self):
        # logging.info("Shutting down listener...")
        if self.listensocket is not None:
            self.listensocket.close()

    def listenthread_secure(self, client_socket, client_address):
        # logging.info("Recv Message from" + str(client_address[0]))

        prv, pub = dq.get_rsa_hex_keys()
        # logging.info("Recv Message, make secure link:")
        scl = SecureCommunicationListen(client_socket, RSA.importKey(bytes.fromhex(prv)), get_configuration_value('listen_read_bytes'))
        if not scl.is_active():
            # logging.info('Listener could not setup secure link with caller, bailing out...')
            return
        # print('Secure link created')
        n = scl.recv()

        # logging.info("Recv Message done.")
        if n:
            # logging.info('Received msg, size=' + str(len(n)))
            pass
        else:
            # logging.info('Recved nothing, close socket and bail out.')
            return
        # logging.info('10-part:' + str(n[:10]))

        # All requests take up 10 bytes
        request = n[:10]
        data = n[10:]
        # print('REQ:', request)
        # print('MSG:', data)
        try:
            # What does he want?

            if request == b'AREUALIVE?':  # Just check if we are a Tenfingers node
                # logging.info('Listen thread got AREUALIVE req.')
                scl.send(b'YES')

            # This is the download data request
            if request == b'ONEFINGER:':  # Rest of the request is a name + RSA_PubKey. If the name corresponds in the database, send back the encrypted file.
                # logging.info('Listen thread got ONEFINGER req.')
                data = data.decode()
                datalist = data.split(',')  # todo: This should just be a name, don't forget to escape it!

                if len(datalist) == 2:
                    # print('datalist:', datalist)
                    name = datalist[0]
                    rsa_pub = datalist[1]
                    # logging.info('Request for: [' + name + ']')

                    # TODO, Serve the name or an error!

                    row = dq.get_link_data_by_name_and_rsapub(name, rsa_pub)

                    if row is not None:
                        # logging.info('Found [' + name + '], points on file [' + row[0] + ']')
                        # Get data from DB
                        file = row[0]
                        data = None

                        # get the folder that is in the /data/ folder (based on the table 'folder' / or our own local folder 'local'):
                        filefolder = os.path.join(locator.create_path('data/') + dq.get_folder(rsa_pub))

                        # logging.info('Trying to serve from folder:' + filefolder)

                        with open(filefolder + "/" + file + '.aes', 'rb') as f:
                            data = f.read()
                        if data is None:
                            # logging.info('[' + name + '] does not point on a file or it was not readable.')
                            scl.send(b'FileNotFound')
                        else:
                            # logging.info('[' + name + '] corresponds to a file and we loaded it, lets send it back:')

                            # Send it back (maybe we need to convert the data, it is raw bytes now)
                            scl.send(data)
                    else:
                        # logging.info('[' + name + '] not found in database.' + rsa_pub)
                        scl.send(b'NotFound')
                else:
                    # logging.info('Malformed request.')
                    scl.send(b'BadRequest')

            # Update request

            if request == b'UPDATEDATA':  # another node asks us to update data we store for them
                logging.info('Listen thread got UPDATEDATA req.')

                filesize = unpack('>Q', data[:8])[0]  # todo: always use data[:8], so a ',' in the pack(...) won't break things! SO fix it everywhere
                # print('Update data REQ, filesize', filesize)

                arr = data[8:].split(b',')  # todo: always use data[8:], so a ',' in the pack(...) won't break things! SO fix it everywhere

                other_name = arr[1].decode()
                other_pub_key_bin_hex = arr[2]

                # print('Verify connecting node...')

                # First verify the caller is actually the owner of this public rsa key:
                random_checksum = b'Digest_this!'+str(random.randint(10000,99999)).encode()+str(time.time()).encode()
                # print('Send random verification data:', random_checksum)
                verification_digest = scl.send(random_checksum)  # caller should now have made a digest off of this checksum
                # print('Got verification digest:', verification_digest)

                # print('Import other pub key:', other_pub_key_bin_hex[:80], '...')

                other_pub_key = RSA.importKey(bytes.fromhex(other_pub_key_bin_hex.decode()))
                # print('Key type:', type(other_pub_key))

                # print('Check digest:')
                digest_ok = rh.verify_digest(random_checksum, other_pub_key, verification_digest)

                if digest_ok:
                    # print('Digest OK, lets check data size:', filesize)
                    # print('other_name type:', type(other_name))
                    logging.info('Got update request for file '+other_name+' of size:'+str(filesize))

                    success = True
                    if filesize > get_configuration_value('swicharoo_file_maxsize_kb')*1024:
                        scl.send(b'Too Big!')
                        success = False

                    if success:
                        # Check we have the place to store the file:
                        used_space = dq.get_shared_files_size()
                        if used_space + filesize > get_configuration_value('max_storage_space_other_mb')*1024*1024:
                            scl.send(b'Out of storage space, sorry.')
                            success = False

                    # TODO: Check the swicharoo doesn't break the swicharoo rules according to the original shared file!
                    # TODO Or else a malicous user can swicharoo a small file, then update it to a much bigger file.
                    if success:
                        pass
                        # Check our swicharoo policy!

                    if success:
                        # print('File size OK.')
                        # todo: verify we actually store this file for this node!

                        other_pub_hex = other_pub_key_bin_hex.decode()
                        # print('other_pub_hex:', other_pub_hex[:32], '...')
                        # print('type:', type(other_pub_hex))
                        if dq.get_link_data_by_name_and_rsapub(other_name, other_pub_hex):
                            # print('Update request: all good!')
                            new_file_data = scl.send(b'ACCEPTING!')

                            # Save other nodes data
                            filefolder = os.path.join(locator.create_path('data/') , dq.get_folder(other_pub_hex))
                            if not os.path.exists(filefolder):
                                os.makedirs(filefolder)

                            save_name = os.path.join(filefolder , other_name + '.aes')
                            logging.info('Save off other file data: ' + save_name)
                            with open(save_name, 'wb') as f:
                                f.write(new_file_data)

                            scl.send(b'OK')
                        else:
                            # print('Update data: we do not store that data here')
                            scl.send(b'DataRemoved')

                else:
                    scl.send(b'UNDIGESTABLE')
                    # print('Update data: digest verification failed')

            # The classic swicharoo:
            if request == b'SWICHAROO?':  # The good old swicharoo

                # We just got a request for swicharoo with a file size and a pub key
                data_size = data[:8]  # Get the first 8 bytes, who represents the others files size
                tmp = data[8:]  # Remove those first 8 bytes, the rest should be the other_pub_key,other_name
                split = tmp.decode().split('|')
                if len(split) != 4:
                    print("#Listener Error in split size, expected five got", len(split))
                    client_socket.close()
                    logging.info('Listener: Got badly formed SWICHAROO string')
                    return

                # We now have all the information about the switch, from their side.
                other_pub_key = split[0]
                other_name = split[1]
                other_ip = split[2]
                other_port = split[3]
                other_file_size = unpack('>Q', data_size)[0]

                # Find the data we want to send back, but do not select data that the other node already store:
                data_id = dq.get_dataid_needing_finger(other_file_size, other_pub_key, other_name)

                logging.info('Listener: Got a Swicharoo request, other_file_size: ' + str(other_file_size) + ' file name: ' + other_name)
                print('#Listener data_id:', data_id)
                if data_id > 0:
                    my_name, my_origfilename = dq.get_link_data_by_dataid(data_id)
                    my_file_size = os.path.getsize(os.path.join(locator.create_path('data/local/'), (my_name + '.aes')))

                    if not self.accept_swicharoo(my_file_size, other_file_size):
                        pass

                    random_checksum = b'Digest_this!' + str(random.randint(10000, 99999)).encode() + str(time.time()).encode()
                    query = b'YES_PLEASE' + pack('>Q', my_file_size) + my_name.encode() + b'|' + my_origfilename.encode() + b'|' + random_checksum
                    #   <<        Send Recv           >>
                    verification_digest = scl.send(query)

                    # Check all the different things that have to work out if we should go further:

                    if verification_digest == b'U_TOO_BIG!':
                        print('They declined our acceptance of their swicharoo request because our file is too big.', flush=True)
                        client_socket.close()
                        logging.info('Listener: Swicharoo turned down our file, too big')
                        return

                    if verification_digest == b'DATA_AGAIN':
                        logging.info('Listener: They seems to already store our data.')
                        client_socket.close()
                        return

                    digest_ok = rh.verify_digest(random_checksum, RSA.importKey(bytes.fromhex(other_pub_key)), verification_digest)
                    if not digest_ok:
                        scl.send(b'DIGEST_ERR')
                        logging.info('Listener: Caller could not authentify themselves. Bail out.')
                        client_socket.close()
                        return

                    if dq.get_link_data_by_name_and_rsapub(other_name, other_pub_key) is not None:  # We already store the data
                        scl.send(b'ALREADY_DO')
                        logging.info('Listener: We already store their data.')
                        client_socket.close()
                        return

                    # Check if we know this node, and if we don't, insert it into the database
                    other_node_id = dq.get_nodeid(other_ip, other_port, other_pub_key)
                    if other_node_id == 0:  # It didn't exist in our database, now it does.
                        dq.insert_node(other_ip, other_port, other_pub_key)

                    # Load up our data so we can send it away:
                    with open(os.path.join(locator.create_path('data/local/'), (my_name + '.aes')), 'rb') as f:
                        file_data = f.read()
                    if file_data is None:
                        logging.info('Listener: Owned data is missing: data/local/' + my_name + '.aes')
                        scl.close()
                        return

                    # All seems okay, proceed with the exchange and send them our data
                    #   <<        Send Recv           >>
                    other_data = scl.send(file_data)

                    # Save other nodes data
                    file_folder = os.path.join(locator.create_path('data/'), dq.get_folder(other_pub_key))
                    if not os.path.exists(file_folder):
                        os.makedirs(file_folder)

                    save_name = os.path.join(file_folder, other_name + '.aes')
                    print('Listener: save theirs:', save_name, flush=True)
                    logging.info('Listener: Save off other file data: ' + save_name)
                    with open(save_name, 'wb') as f:
                        f.write(other_data)

                    # Insert a pointer to our data (their node)
                    dq.insert_finger(other_node_id, data_id)

                    # DONE!
                    scl.send(b'OK')
                else:
                    scl.send(b'SRYNODATA.')
                    logging.info('Listener: Swicharoo request turned down because we do not have anything to share that the offer corresponds to')



            if request == b'SEND_NODES':  # send back a random node for now  # todo: send back several
                # check if we already have this node, and if not, add it:
                data = data.decode()
                datalist = data.split(',')  # This should be IP/PORT data from caller todo: Escape it!
                if len(datalist) == 2:
                    port = datalist[0]
                    rsa_pub = datalist[1]

                    # This call will silently fail if the node already exists
                    dq.insert_node(client_address[0], port, rsa_pub)

                else:
                    logging.info('#Warning: SEND_NODES req malformed, len(datalist) =' + str(len(datalist)))


                nodeid = dq.get_random_node_id(False)  # todo: remove the False from the call so we send back only valid nodes
                if nodeid is not None:
                    tmp = dq.get_node_data(nodeid)
                    j = {'ip': tmp[0], 'port': tmp[1], 'rsa_pub_key': tmp[2]}
                    #print('Send back node:', tmp[0], tmp[1], tmp[2][:22], flush=True)
                    # print('j:', j)
                    try:
                        r = json.dumps(j)
                    except:
                        print('Could not dumps the dict')
                    # print('dumps=', r, flush=True)
                    b = r.encode()  # transform to byte object
                    # print('send:', b, flush=True)
                    scl.send(b)
                else:
                    scl.send(b'{}')
                    print('We do not have any nodes to share yet!', flush=True)
                #print('SEND NODES Done.', flush=True)

            # Close the socket
            client_socket.close()

            if request == b'NEW_METADT':  # A node is requiring an update of its address list, metadata,...
                tmp = data.split(b',')  # name + ',' + pub + ',' + epoch_time + address_list, digest
                if len(tmp) == 5:
                    name = tmp[0]
                    pub = tmp[1]
                    epoch_time = tmp[2]
                    digest = tmp[3]
                    address_list = tmp[4]  # address list is not used in the digest.

                    # first we verify this public key shows the private key has signed the message:
                    if rh.verify_digest(name + ',' + pub + ',' + epoch_time, pub, digest):
                        # if so, then update the database (as data is stored by name+pub_key, this works)
                        print('NEW_METADT request: verification OK, insert meta data and address list')
                        dq.try_update_address_list(name, pub, address_list, epoch_time)
                    else:
                        print('NEW_METADT request: digest verification failed')
                else:
                    print('Got mal formed NEW_METADT request, incorrect size is:', len(tmp))


        except NameError as error:
            logging.info("[NameError] Message not understood: NameError=" + str(error))
        except:
            logging.info("[General Exception] Message not understood: " + str(sys.exc_info()[0]))
            scl.send(b'GeneralError')
        # logging.info('Listen thread done.')

        if scl.is_active():
            scl.close()

    def run(self):
        print('inner thread start')
        try:
            self.listensocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.listensocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            #self.listensocket.bind((str(socket.INADDR_ANY), self.port))
            self.listensocket.bind(('', get_configuration_value('port')))
            self.listensocket.listen(get_configuration_value('max_simultaneous_connections'))
            self.listensocket.settimeout(get_configuration_value('listen_timeout_seconds'))

        except socket.error as err:
            print('Failed to create listen socket: ', err)
            logging.info('Failed to create listen socket')
            exit()

        # log
        logging.info('Listen on ' + get_configuration_value('ip') + ':' + str(get_configuration_value('port')))

        global force_shutdown
        while not force_shutdown:

            gotconnection = False
            while not force_shutdown and gotconnection is False:
                try:
                    # establish connection
                    client_socket, client_address = self.listensocket.accept()
                    # Pass through the while:
                    gotconnection = True
                    # Make a new thread that will deal with the communication:
                    if not force_shutdown:
                        # Read request:
                        client_socket.settimeout(60)
                        # Start listener
                        threading.Thread(target=self.listenthread_secure, args=(client_socket, client_address)).start()

                except socket.error:  # as err:
                    # print("listen exception", err)
                    pass


version = '1.07'

logging.info('Tenfingers listener startup: V' + version + ' start time: ' + str(datetime.datetime.now()))
logging.info('Use --noscheduler to disable scheduler')
logging.info('Listen configuration: ' + get_configuration_value('ip') + ':' + str(get_configuration_value('port')))

run_scheduler = True
if len(sys.argv) > 1 and sys.argv[1] == '--noscheduler':
    logging.info('Scheduler will not be started (--noscheduler)')
    run_scheduler = False



time.sleep(0.1) # So logging will have time to write

# set up a thread with the scheduler
from scheduler import Scheduler
scheduler = None
scheduler_thread = None
if run_scheduler:
    scheduler = Scheduler()
    scheduler_thread = threading.Thread(name='Scheduler', target=scheduler.run)
    scheduler_thread.start()
    time.sleep(0.1)  # So logging will have time to write

# Create thread and start up:
listener = Listen()
listener_thread = threading.Thread(name='Listener', target=listener.run)  # l.run2, args=("abc",))

logging.info("start listener thread:")
listener_thread.start()


# Set up a function that will be called at SIGINT so that a CTRL+C will shut down all threads and exit correctly:
def signal_handler(sig, frame):
        logging.info('Received SIGINT, shutting down threads.')
        global force_shutdown
        force_shutdown = True
        listener.forcestop()


# Set up the SIGINT handler
signal.signal(signal.SIGINT, signal_handler)
