# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

from Crypto.PublicKey import RSA
from Crypto import Random

from Crypto.Cipher import PKCS1_OAEP
from Crypto.Signature import PKCS1_v1_5
import hashlib

from Crypto.Util import Counter

from binascii import hexlify

import database_query as dq


# Saves a RSA key file as an encoded file
def save_key(filename, key):
    try:
        f = open(filename, 'wb')
        f.write(key.exportKey('DER'))
        f.close()
    except IOError as err:
        print("Exception raised at save key file", err)
        return False
    return True


# Loads a RSA key file from an encoded file
def load_key(filename):
    try:
        with open(filename, 'rb') as content_file:
            content = content_file.read()
            return RSA.importKey(content)
    except IOError as err:
        print("Exception raised at load key file", err)
    return None


# Gets the unique RSA keypair from the database or None if not existing. This function returns fully functional keys!
def get_rsa_keypair(noerror=False):  # Returns prv, pub
    prv, pub = dq.get_rsa_hex_keys()
    if prv is not None:
        return RSA.importKey(bytes.fromhex(prv)), RSA.importKey(bytes.fromhex(pub))
    else:
        if noerror:
            return None
        print('#Error: RSA Keypair not found in database!')
        exit(-7)


#


# Generates a RSA keypair
def generate_rsa_keypair(keysize):
    print("Generating RSA Key pair, please wait...")
    random_generator = Random.new().read
    prv = RSA.generate(keysize, random_generator)
    # prv = RSA.generate(keysize, 65537, callback=lambda x, y, z: None)
    pub = prv.publickey()
    return prv, pub


def pub_encrypt(self, pubkey, message):
    cipher = PKCS1_OAEP.new(pubkey)
    return cipher.encrypt(message)


def prv_decrypt(self, prvkey, message):
    cipher = PKCS1_OAEP.new(prvkey)
    return cipher.decrypt(message)


# Returns a short byte string
def _sha(message):
    if not isinstance(message, bytes):
        message = message.encode()
    return hashlib.sha256(message).hexdigest().encode()


# creates a digest, extracts the number and returns a hex version of it (of type bytes so we can easily send it over the network)
def create_digest(message, rsa_prv_key):
    if not isinstance(message, bytes):
        print('Warning: create_digest was called with non-bytes variable: auto encodes it to bytes')
        message = message.encode()
    message = _sha(message)
    d = rsa_prv_key.sign(message, '')
    d0 = d[0]
    h = hex(d0)
    b = h.encode()
    return b


# digest must be in hex-numbers byte form
def verify_digest(message, rsa_pub_key, digest):
    # Check if the key id str, bytes, real key (we need a real key)
    message = _sha(message)
    h = digest.decode()
    d0 = int(h, 16)
    d = (d0,)
    return rsa_pub_key.verify(message, d)


# use this to validate / create digests for addresses
def _make_address_string(ip, port, pub_key):  # Just concats the data so we can use create digest and verify easily
    if not isinstance(ip, bytes):
        ip = ip.encode()
    if not isinstance(port, bytes):
        port = port.encode()
    if not isinstance(pub_key, bytes):
        pub_key = pub_key.encode()
    return ip + b'_' + port + b'_' + pub_key + b'_end'


# Use this to create an universal digest for an address
def create_address_digest(rsa_prv_key, ip, port, pub_key):
    return create_digest(_make_address_string(ip, port, pub_key), rsa_prv_key)


# Use this to validate an address
def verify_address_digest(bmessage, digest, pub_key):
    return verify_digest(bmessage, pub_key, digest)

















